package fr.cnam.poker.serveur;

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.Inet4Address;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import fr.cnam.poker.commun.Carte;
import fr.cnam.poker.commun.Combinaison;
import fr.cnam.poker.commun.Joueur;
import fr.cnam.poker.commun.TablePoker;
import fr.cnam.poker.robot.FactoryRobot;

// Classe de dfinition du MODELE du jeu de Poker
// 
// Cette classe est un SINGLETON car le MODELE est unique sur le serveur
//
public class ModeleJeuPoker {

	// Dfinition du SINGLETON
	private static ModeleJeuPoker modeleJeuPoker=null;
	public static ModeleJeuPoker getInstance() {
		if(modeleJeuPoker==null)modeleJeuPoker=new ModeleJeuPoker();
		return modeleJeuPoker;
	}

	private ObservablePoker observable;  // L'Observable du Modle permettant de faire la notification  toutes les IHM
	private TablePoker tablePoker;       // La table de Poker qui contient tous les lments  notifier
	private Sabot sabot;                 // Le sabot qui contient toutes les cartes (distribues et restantes)
	private boolean premiereFois;        // Vrai si le jeu n'a pas commenc sinon Faux. Permet d'initialiser
	// le premier donneur et de mlanger le jeu de carte

	// Constructeur du MODELE
	//
	private ModeleJeuPoker() {
		this.tablePoker = new TablePoker();
		this.observable = new ObservablePoker();
		this.sabot = new Sabot();
		this.premiereFois=true;
	}

	// Distribuer les cartes  chacun des joueurs
	//
	public void distribuer() {
		//TODO : Distribuer en commencant par celui  gauche du donneur
		for(Joueur joueur:tablePoker.getJoueurs()) {
			if (!joueur.isAbandon()) {
				joueur.getCartes()[0]=sabot.prendreCarte();
				joueur.getCartes()[1]=sabot.prendreCarte();
			}
		}
	}

	// Abonnement d'un joueur qui donne ses informations de connexion : son host et son port
	//  Cette mthode est appel dans le CONTROLEUR sur la requte d'abonnement
	//
	//  Retourne -1 si le nombre de joueurs est superieur  4
	//
	public int sabonner(String host,int port,String nomJoueur) {
		if (tablePoker.getNbJoueur()>=4) return -1;

		ObserverPokerIHM observer = new ObserverPokerIHM(host,port);
		observable.addObserver(observer);
		Joueur j = new Joueur();
		j.setAbandon(false);
		j.setNom(nomJoueur);
		if (tablePoker.getNbJoueur()==0) j.setDonneur(true); // Le premier joueur est le donneur
		tablePoker.addJoueur(j);  // Ajout du joueur
		// Si le joueur est un Robot, on le cre
		// Un joueur est un robot si son nom commence par "ROBOT"
		if (nomJoueur.startsWith("ROBOT")) {
			j.setRobot(true);
			FactoryRobot.getInstance().getRobot(j); // Cration du robot associ au joueur
		}
		// 
		return j.getNumero(); // Retourne l'indice du joueur dans le tableau dejoueur
		// Cette valeur est envoy au client pour qu'il connaisse son numro de joueur

	}

	// Mthode qui ralise l'action d'un joueur
	// L'action est dfini par :
	//    - le numro du joueur qui a ralis l'action
	//    - le nom de l'action ralis
	//    - les paramtres ventuels de l'action (tableau de String)
	//
	public void actionJoueur(int numeroJoueur,String action,String... parametres)
	{
		String s=" ";for(String p:parametres)s=s+" "+p;
		if (action.equals("ENVOYER_MESSAGE"))
			tablePoker.addInfo(tablePoker.getJoueur(numeroJoueur).getNom()+"> "+s);
		else
			tablePoker.addInfo(tablePoker.getJoueur(numeroJoueur).getNom()+"> "+action+s);


		if(action.equals("COMMENCER")) {
			commencerUneManche(numeroJoueur);
			passerAuJoueurSuivant();
		}

		if(action.equals("MISER")) {
			Joueur joueur = tablePoker.getJoueur(numeroJoueur);
			int mises = Integer.parseInt(parametres[0]);
			joueur.miser(mises);
			passerAuJoueurSuivant();
		}
		if(action.equals("CHANGER_BLIND")) {
			tablePoker.setBlind(Integer.parseInt(parametres[0]));
		}
		if(action.equals("MISER_BLIND")) {
			Joueur joueur = tablePoker.getJoueur(numeroJoueur);
			joueur.miser(tablePoker.getBlind());
			passerAuJoueurSuivant();
		}
		if(action.equals("MISER_DOUBLE_BLIND")) {
			Joueur joueur = tablePoker.getJoueur(numeroJoueur);
			joueur.miser(tablePoker.getBlind()*2);
			passerAuJoueurSuivant();
		}
		if(action.equals("FLOP")) {
			sabot.prendreCarte(); // On brule la carte
			tablePoker.getFlop()[0]=sabot.prendreCarte();
			tablePoker.getFlop()[1]=sabot.prendreCarte();
			tablePoker.getFlop()[2]=sabot.prendreCarte();
			calculerMainDesJoueurs(); // HORS SUJET
		}
		if(action.equals("TURN")) {
			sabot.prendreCarte(); // On brule la carte
			tablePoker.setTurn(sabot.prendreCarte());
			calculerMainDesJoueurs(); // HORS SUJET
		}
		if(action.equals("RIVER")) {
			sabot.prendreCarte(); // On brule la carte
			tablePoker.setRiver(sabot.prendreCarte());
			calculerMainDesJoueurs(); // HORS SUJET
		}
		if(action.equals("CHECKER")) {
			passerAuJoueurSuivant();
		}
		if(action.equals("ABANDONNER")) {
			tablePoker.getJoueur(numeroJoueur).setAbandon(true);
			passerAuJoueurSuivant();
		}
		if(action.equals("RAMASSER_POT")) {
			ramasserPot(numeroJoueur);
		}
		if(action.equals("FAIRE_TAPIS")) {
			tablePoker.getJoueur(numeroJoueur).setTapis(true);
			passerAuJoueurSuivant();
		}
		if(action.equals("MONTRER_CARTES")) {
			tablePoker.getJoueur(numeroJoueur).setCartes_visibles(true);
			// Si au moins 1 robot montrer sescartes entraine de voir aussi celles du robot
			if (tablePoker.auMoinsUnRobot()) for(Joueur j:tablePoker.getJoueurs()) j.setCartes_visibles(true);
		}
		if(action.equals("GAGNANT")) {
			if (tablePoker.getFlop()[0]==null)tablePoker.addInfo(tablePoker.getJoueur(numeroJoueur).getNom()+">Action GAGNANT impossible");
			else determinerGagnant();
		}

		// Dans tous les cas, suite  une action d'un joueur, on notifie  toutes les IHM
		//   le nouvel tat de la table de poker
		//
		observable.notifier(tablePoker);
	}

	// Mthode qui permet de commencer une mmanche (la 1re ou une nouvelle)
	//   numeroJoueur : le numro du joueur qui commence  donner
	//
	private void commencerUneManche(int numeroJoueur) {
		if (premiereFois) {   // Si c'est la 1re manche
			tablePoker.setCommence(true);
			sabot.melanger();
			premiereFois=false;	
			// La premre fois (le jeu n'a pas commenc) le donneur est le joueur qui a 
			//   ralis l'action de commencer
			tablePoker.getJoueur(numeroJoueur).setDonneur(true);

		}
		else {  // On commence une nouvelle manche

			// Initialisation des joueurs
			for(Joueur joueur:tablePoker.getJoueurs()) joueur.initManche();

			// Dterminer le donneur suivant 
			int nouveauDonneur = calculerDonneurSuivant(numeroJoueur);
			tablePoker.getJoueur(numeroJoueur).setDonneur(false);
			tablePoker.getJoueur(nouveauDonneur).setDonneur(true);
			// Le joueur courant est donc le nouveau donneur (meme si plus loin) le joueur courant
			//  va devenir le joueur suivant du donneur
			tablePoker.setNumeroJoueurCourant(nouveauDonneur);

		}
		// Initialisation de la manche
		//
		tablePoker.getFlop()[0]=null;
		tablePoker.getFlop()[1]=null;
		tablePoker.getFlop()[2]=null;
		tablePoker.setTurn(null);
		tablePoker.setRiver(null);
		sabot.remettreCartes();
		sabot.couper();
		distribuer();
	}

	// Mthode qui permet de dterminer si au moins 1 joueur  fait tapis.
	//  Si un joueur au moins a fait tapis alors on ramasse les jetons des autres joueurs (les perdants)
	//   en fonction des jetons du joueur gagnant
	//
	// Cette mthode retourne vrai s'il y eu au moins 1 joueurs qui a fait tapis
	//
	// numeroJoueur = numro du joueur qui ramasse le pot (gagnant)
	//
	private boolean ramasserTapis(int numeroJoueur){
		boolean existeTapis=false;
		Joueur joueur = tablePoker.getJoueur(numeroJoueur);
		int tapisJoueur = joueur.getNbJetonsMises()+joueur.getNbJetons();
		int gainJoueur=0;
		for(Joueur adversaire:tablePoker.getJoueurs()) {
			if (adversaire.isTapis() && adversaire.getNumero()!=numeroJoueur) {
				existeTapis=true;

				int tapisAdversaire = adversaire.tousLesJetons();
				if (tapisAdversaire<tapisJoueur) {
					adversaire.setNbJetons(0);
					gainJoueur=gainJoueur+tapisAdversaire;
				}
				else {
					adversaire.setNbJetons(tapisAdversaire-tapisJoueur);
					gainJoueur=gainJoueur+tapisJoueur;
				}
				adversaire.setNbJetonsMises(0);	
			}
		}
		joueur.ajouterJetons(gainJoueur+joueur.getNbJetonsMises());
		joueur.setNbJetonsMises(0);
		return existeTapis;
	}

	// Mthode qui est excute quand un joueur a dcid de rammasser le pot (il a gagn)
	// Il y a 2 cas de figure :
	//    - au moins 1 joueur a fait tapis : le gagnant ramasse le pot et les jetons des autres joueurs
	//    - aucun joueur n'a fait tapis : le gagnant ne ramasse que le pot
	//
	private void ramasserPot(int numeroJoueur) {
		boolean existeTapis = ramasserTapis(numeroJoueur);
		if (!existeTapis) {
			int somme=0;
			for(Joueur joueur:tablePoker.getJoueurs()) {
				somme=somme+joueur.getNbJetonsMises();
				joueur.setNbJetonsMises(0);
			}
			int jetons=tablePoker.getJoueur(numeroJoueur).getNbJetons();
			tablePoker.getJoueur(numeroJoueur).setNbJetons(jetons+somme);
		}
	}

	// Mthode qui permet de passer a joueur suivant
	//  
	public void passerAuJoueurSuivant() {
		int numeroJoueurCourant = tablePoker.getNumeroJoueurCourant();
		int nouveauNumeroJoueurCourant=calculerJoueurSuivant(numeroJoueurCourant);
		tablePoker.setNumeroJoueurCourant(nouveauNumeroJoueurCourant);

		// HORS SUJET : Si le joueur suivant est un robot alors il joue
		//
		if(tablePoker.getJoueur(nouveauNumeroJoueurCourant).isRobot()) {
			FactoryRobot.getInstance().getRobot(tablePoker.getJoueur(nouveauNumeroJoueurCourant)).jouer();
		}
	}

	// Mthode qui calcule le joueur suivant d'un joueur
	//  Il faut sauter les joueurs qui ont abandonns
	//
	public int calculerJoueurSuivant(int numeroJoueur) {
		int suivant=(numeroJoueur+1)%tablePoker.getNbJoueur();
		while(suivant!=numeroJoueur) {
			Joueur joueur = tablePoker.getJoueur(suivant);
			if (!joueur.isAbandon()) return suivant;
			else suivant=(suivant+1)%tablePoker.getNbJoueur();
		}
		return suivant;
	}

	// Mthode qui calcule le donneur suivant
	//  Il faut sauter les robots qui ne peuvent pas tre des donneurs (HORS SUJET)
	//
	public int calculerDonneurSuivant(int numeroJoueur) {
		int suivant=(numeroJoueur+1)%tablePoker.getNbJoueur();
		while(suivant!=numeroJoueur) {
			Joueur joueur = tablePoker.getJoueur(suivant);
			if (!joueur.isRobot()) return suivant;
			else suivant=(suivant+1)%tablePoker.getNbJoueur();
		}
		return suivant;
	}



	// Mthode lmentaire d'envoi d'une requte  un host/port
	// Cette mthode est utilise par le serveur pour envoyer une requte  une des IHM (host/port)
	// Elle est utilise pour faire la notification aux IHMs
	//
	public Object envoyerRequete(String hostIHMPoker,int portIHMPoker,Object obj)  {
		try {
			Socket soc = new Socket(Inet4Address.getByName(hostIHMPoker),portIHMPoker);
			OutputStream os=soc.getOutputStream();
			ObjectOutputStream oos=new ObjectOutputStream(os);
			InputStream is=soc.getInputStream();
			ObjectInputStream ois=new ObjectInputStream(is);


			oos.writeObject(obj);
			Object retour=ois.readObject();
			soc.close();
			return retour;
		}catch(Exception ex) {ex.printStackTrace();return null;}
	}


	public ObservablePoker getObservable() {
		return observable;
	}

	public TablePoker getTablePoker() {
		return tablePoker;
	}

	// Calculer la main de poker qu' chacun des joueurs
	//
	public void calculerMainDesJoueurs() {
		for(Joueur joueur:tablePoker.getJoueurs()) {
			ArrayList<Carte> cartes = new ArrayList<Carte>();
			cartes.add(joueur.getCartes()[0]);
			cartes.add(joueur.getCartes()[1]);
			cartes.add(ModeleJeuPoker.getInstance().getTablePoker().getFlop()[0]);
			cartes.add(ModeleJeuPoker.getInstance().getTablePoker().getFlop()[1]);
			cartes.add(ModeleJeuPoker.getInstance().getTablePoker().getFlop()[2]);
			if (ModeleJeuPoker.getInstance().getTablePoker().getTurn()!=null)
				cartes.add(ModeleJeuPoker.getInstance().getTablePoker().getTurn());
			if (ModeleJeuPoker.getInstance().getTablePoker().getRiver()!=null)
				cartes.add(ModeleJeuPoker.getInstance().getTablePoker().getRiver());

			Combinaison comb=Combinaison.calculerMainDe5cartesParmiCartes(cartes);
			joueur.setCombinaison(comb);
			System.out.println(comb);
		}
	}

	// Permet de calculer qui a gagn ou qui sont  galits en fonction de
	//   leurs mains de poker
	//
	public void determinerGagnant() {
		
		// On cre la liste des joueurs qui n'ont pas abandonns 
		//
		ArrayList<Joueur> joueursTrieParCombinaison=new ArrayList<Joueur>();
		for(Joueur j:tablePoker.getJoueurs())
			if (! j.isAbandon()) {
				joueursTrieParCombinaison.add(j);
				tablePoker.addInfo(j.getNom()+" : "+Combinaison.MAINS[j.getCombinaison().getValeurMain()]);
			}
		if (joueursTrieParCombinaison.size()==0) tablePoker.addInfo("Tous les joueurs ont abandonns!!");
		if (joueursTrieParCombinaison.size()==1) tablePoker.addInfo("Il ne reste qu'1 seul joueur : "+joueursTrieParCombinaison.get(0).getNom());
		
		// Il y a au moins 2 joueurs : on trie par ordre croissant de leurs combinaisons
		//
		Collections.sort(joueursTrieParCombinaison,new ComparerJoueurParCombinaison());
		
		// On dtermine les joueurs qui seraient gaux en haut de la liste
		//
		ArrayList<Joueur> joueursEgaux=new ArrayList<Joueur>();
		joueursEgaux.add(joueursTrieParCombinaison.get(joueursTrieParCombinaison.size()-1));
		for(int k=joueursTrieParCombinaison.size()-1;k>0;k--) {
			int res=Combinaison.comparer2Combinaisons(joueursTrieParCombinaison.get(k).getCombinaison(),
					                                  joueursTrieParCombinaison.get(k-1).getCombinaison());
			if (res==0) joueursEgaux.add(joueursTrieParCombinaison.get(k-1));
		}
		
		// Si un seul dans la liste alors il est le gagant
		if( joueursEgaux.size()==1) tablePoker.addInfo("GAGNANT = "+joueursEgaux.get(0).getNom());
		// Il y a au moins 2 gaux
		else {
			String s="";
			for(Joueur j:joueursEgaux) s=s+j.getNom()+" ";
			tablePoker.addInfo(joueursEgaux.size()+" joueurs  galits : "+s);
		}
	}

	class ComparerJoueurParCombinaison implements Comparator<Joueur>{
		@Override
		public int compare(Joueur o1, Joueur o2) {
			// TODO Auto-generated method stub
			return Combinaison.comparer2Combinaisons(o1.getCombinaison(), o2.getCombinaison());
		}
	}
}
